function PUGEBase() {
  this.user = null;
  this.notifiedEntries = {};

  this.fixtures = null;
  this.matches = {};
  this.notifiee = null;
  this.playHistory = null;
  this.leaderboard = null;

  this.loginAttempts = Config.loginAttempts;

  var apiBase = Config.pugeServer + "api/v1/soccer/";
  this.api = {
    fixtures: apiBase + 'round/pl/current.json',
    user: apiBase + 'user/ping.json',
    register: apiBase + 'user/create.json',
    resendActivation: apiBase + 'user/resend_activation.json',
    addLogin: apiBase + 'user/add_login.json',
    entry: {
      create: apiBase + 'entry/create.json'
    },
    match: function(id) {
      return apiBase + 'match/' + id + '.json';
    },
    redirector: Config.pugeServer + 'redirector',
    playHistory: apiBase + 'play_history.json',
    leaderboard: apiBase + 'rankings/round.json'
  };

  this._setCredentials({
    login: localStorage.getItem('login'),
    password: localStorage.getItem('password'),
    facebook_user_id: localStorage.getItem('facebook_user_id'),
    facebook_hash: localStorage.getItem('facebook_hash')
  });
}

PUGEBase.prototype.notify = function(eventName, extra1) {
  if (this.notifiee && eventName in this.notifiee) {
    this.notifiee[eventName](extra1);
  }
};

PUGEBase.prototype._forNewOutcomes = function(callback) {
  if (this.user) {
    for (var gameId in this.user.entriesWithOutcomes) {
      var entries = this.user.entriesWithOutcomes[gameId];

      for (var i in entries) {
        var entry = entries[i];
        if (!this.notifiedEntries[entry.id]) {
          if (callback) {
            callback(entry);
          }

          this.notifiedEntries[entry.id] = true;
        }
      }
    }
  }
};

PUGEBase.prototype.refreshMatch = function(id) {
  var self = this;

  $.ajax({
    type: 'GET',
    url: this.api.match(id),
    dataType: 'json',
    success: function(newMatch, status, xhr) {
      UI.disconnected(false);

      if (!objectEquals(self.matches[newMatch.id], newMatch)) {
        self.matches[newMatch.id] = objectBecome(newMatch, Match);
        self._assignHeaderDateToMatches([newMatch], xhr);
        self.notify('matchUpdated', newMatch);
      } 
    },
    error: function(response) {
      UI.disconnected(true);
      console.log("Unable to load fixture content: " + response.status + "<br/>" + response.responseText);
      return false;
    }
  });  
};

PUGEBase.prototype._copyPlayersFromMatches = function() {
  for (var i in this.fixtures) {
    var match = this.fixtures[i];
    var bigMatch = this.matches[match.id];

    if (bigMatch) {
      match.usePlayersFrom(bigMatch);
    }
  }
};

PUGEBase.prototype._assignHeaderDateToMatches = function(matches, xhr) {
  var date = Util.extractDateFromHeaders(xhr.getAllResponseHeaders());

  for (var i in matches) {
    matches[i].setNow(date);
  }
};

PUGEBase.prototype.refreshFixtures = function() {
  var self = this;

  $.ajax({
    type: 'GET',
    url: this.api.fixtures,
    dataType: 'json',
    success: function(newFixtures, status, xhr) {
      UI.disconnected(false);

      self.fixtures = eachBecome(newFixtures, Match);
      self._assignHeaderDateToMatches(newFixtures, xhr);
      self._copyPlayersFromMatches();
      self.notify('fixturesUpdated');
    },
    error: function(response) {
      UI.disconnected(true);
      console.log("Unable to load fixture content: " + response.status + "\n" + response.responseText);
      return false;
    }
  });
};

PUGEBase.prototype.refreshPlayHistory = function(page) {
  var self = this;
  $.ajax({
    type: 'POST',
    url: this.api.playHistory + "?page=" + this._gotoPage(page),
    data: this.credentials,
    dataType: 'json',
    success: function(newPlayHistory, status, xhr) {
      self.playHistory = eachBecome(newPlayHistory, PlayHistory);
      self.notify("playHistoryUpdated", self.playHistory);
    },
    error: function(response) {
      UI.disconnected(true);
      console.log("Unable to load Play History content: " + response.status + "\n" + response.responseText);
      return false;
    }
  });
};

PUGEBase.prototype._gotoPage = function(page) {
  return (page || (this.playHistory && this.playHistory.page.current_page) || 1);
};

PUGEBase.prototype.refreshLeaderboard = function() {
  var self = this;
  $.ajax({
    type: 'GET',
    url: this.api.leaderboard,
    dataType: 'json',
    success: function(newLeaderboard, status, xhr) {
      self.leaderboard = eachBecome(newLeaderboard, Leaderboard);
      self.notify("leaderboardUpdated", self.leaderboard);
    },
    error: function(response) {
      UI.disconnected(true);
      console.log("Unable to load Leaderboard content: " + response.status + "\n" + response.responseText);
      return false;
    }
  });
};

PUGEBase.prototype._setCredentials = function(credentials) {
  if (credentials) {
    this.credentials = credentials;
    if(credentials.login && credentials.password) {
      localStorage.setItem('login', credentials.login);
      localStorage.setItem('password', credentials.password);
    } else if (credentials.facebook_user_id && credentials.facebook_hash) {
      localStorage.setItem('facebook_hash', credentials.facebook_hash);
      localStorage.setItem('facebook_user_id', credentials.facebook_user_id);
      localStorage.setItem('access_token', credentials.access_token);
    }
  }
  else {
    this.credentials = {};
    localStorage.removeItem('login');
    localStorage.removeItem('password');
    localStorage.removeItem('facebook_hash');
    localStorage.removeItem('facebook_user_id');
    localStorage.removeItem('access_token');
  }
};

PUGEBase.prototype._setUser = function(user) {
  var userWasNull = (!this.user || !this.user.entriesWithOutcomes);

  this.user = user;
  this.notify('userUpdated');

  // Note that we still want to run the whole loop, even if we don't supply a callback
  // This will flush the new outcomes so we don't hear about old news
  var self = this;
  this._forNewOutcomes(userWasNull ? undefined : function(entry) {
    self.notify('entryOutcome', entry);
  });
};

// TODO get rid of repetition
PUGEBase.prototype.refreshUser = function() {
  var self = this;

  if (this.user)
  {
    this._checkCredentials(this.credentials, function(user, error, communicationFailure) {
      if (user) {
        self._setUser(objectBecome(user, User));
      }
      else if (!communicationFailure) {   // if there was a communication failure, we don't want to log them out
        self.logout();
      }
    });
  }
};

PUGEBase.prototype.interestingGames = function() {
  var result = [];

  for (var i in this.fixtures) {
    var match = this.fixtures[i];

    if (match.games.soccer1up) {
      result.push(match.games.soccer1up.gameId);
    }
  }

  return result.join('-');
};

PUGEBase.prototype._willHandleError = function(response) {
  return !!{
    423: 'Account not verified',
    422: 'Account login is missing',
    403: 'Forbidden'
  }[response.status];
};

PUGEBase.prototype._checkCredentials = function(credentials, callback) {
  var self = this;

  $.ajax({
    type: 'POST',
    url: this.api.user,
    data: {
      games: this.interestingGames(),
      login: credentials.login,
      password: credentials.password,
      facebook_user_id: credentials.facebook_user_id,
      access_token: credentials.access_token,
      facebook_hash: credentials.facebook_hash
    },
    dataType: 'json',
    success: function(content) {
      UI.disconnected(false);

      if (callback) {
        callback(content);
      }
    },
    error: function(response) {
      if (credentials.facebook_user_id && response.status == 403) {
        credentials.facebook_hash = JSON.parse(response.responseText).facebook_hash;
        if(credentials.facebook_hash !== undefined) {
          // we have a valid facebook hash, run check credentials again to confirm login success
          PUGE._checkCredentials(credentials, callback);
        } else {
          callback(null, Util.createErrorObject(response), false);
        }
      } else if (self._willHandleError(response)) {
        callback(null, Util.createErrorObject(response), false);
      } else {
        UI.disconnected(true);
        UI.closePopup();
        console.log('Unable to check credentials with server: ' + response.status + "\n" + response.responseText);
        callback(null, { error: "Error connecting to server" }, true);
      }
    }
  });
};

PUGEBase.prototype.login = function(credentials, loginCallback) {
  var self = this;

  var callback = function(user, error, communicationFailure) {

    self.loginAttempts--;

    if (user) {
      self.setCredentialsAndUser(credentials, user);
    }
    else if (communicationFailure && self.loginAttempts > 0) {
      self._checkCredentials(credentials, callback);
    }
    else {
      self.logout();
    }

    var loginCompleted = !!user || !self.loginAttempts || (error && !communicationFailure);

    if (loginCompleted && loginCallback) {
        loginCallback(user, error);
    }
  };

  this._checkCredentials(credentials, callback);
};

PUGEBase.prototype.register = function(params, callback) {
  var self = this;
  $.ajax({
    type: 'POST',
    url: this.api.register,
    data: params,
    dataType: 'json',
    success: function(user) {
      UI.disconnected(false);

      callback(true);
    },
    error: function(response) {
      UI.disconnected(true);

      try {
        callback(false, JSON.parse(response.responseText).errors);
      } catch(e) {
        callback(false, ['An unknown error occurred']);
        UI.closePopup();
      }
    }
  });
};

PUGEBase.prototype.setCredentialsAndUser = function(credentials, user) {
  this._setCredentials(credentials);
  this._setUser(objectBecome(user, User));
};

PUGEBase.prototype.logout = function() {
  this._setCredentials(null);
  this._setUser(null);
};

PUGEBase.prototype.placeEntry = function(entry, callback) {
  var self = this;

  if (this.user) {
    $.ajax({
      type: 'POST',
      url: this.api.entry.create,
      data: {
        team_id: entry.team.id,
        scorer_id: entry.player.id,
        game_id: entry.game.gameId,
        type: entry.type,
        interested_games: this.interestingGames(),
        login: this.credentials.login,
        password: this.credentials.password,
        facebook_user_id: this.credentials.facebook_user_id,
        facebook_hash: this.credentials.facebook_hash,
        access_token: this.credentials.access_token
      },
      dataType: 'json',
      success: function(result) {
        UI.disconnected(false);

        if (result.user) {
          self._setUser(objectBecome(result.user, User));
        } else {
          self.user.entries[entry.game.id] = [entry]; // TODO temporary hack, do properly.
        }

        if (callback) {
          callback(true);
        }

        $(".sending").removeClass("sending");
      },
      error: function(xhr, type) {

        if (xhr.status == 403) {
          if (callback) {
            var errorText = "An unknown error occurred";
            var errorTitle = "Message";
            var url;
            if(type == 'error') {
              var responseJson = JSON.parse(xhr.responseText);
              if(responseJson.error) {
                errorText = responseJson.error;
                url = responseJson.url;
                if (responseJson.title) {
                  errorTitle = responseJson.title;
                }
              }
            }
            callback(false, errorTitle, errorText, url);
          }
        } else {
          UI.disconnected(true);
        }
        $(".sending").removeClass("sending");
      }

    });
  }
  else {
    callback(false, "Please login to place an entry.");
  }
};

PUGEBase.prototype.redirectorUrl = function(path, callback) {
  if (this.user) {
    $.ajax({
      type: 'POST',
      url: this.api.redirector,
      data: {
        login: this.credentials.login,
        password: this.credentials.password,
        destination: path
      },
      dataType: 'json',
      success: function(result) {
        UI.disconnected(false);
        
        callback(result.url);
      },
      error: function(xhr) {
        UI.disconnected(true);

        callback(Config.pugeServer + path);
      }
    });
  } else {
    callback(Config.pugeServer + path);
  }
};

PUGE = new PUGEBase();
